package fr.ippon.kafkaLatency;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import fr.ippon.kafkaLatency.model.Person;
import fr.ippon.kafkaLatency.utils.Config;
import fr.ippon.kafkaLatency.utils.MetricsProducerReporter;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.clients.admin.CreateTopicsResult;
import org.apache.kafka.clients.admin.NewTopic;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.errors.TopicExistsException;
import org.apache.kafka.common.serialization.StringSerializer;
import org.apache.kafka.connect.json.JsonSerializer;
import org.apache.log4j.Logger;

import java.beans.IntrospectionException;
import java.io.File;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeoutException;

/**
 * IPPON 2018.
 */
public class AlimKafka {

    private static ClassLoader classLoader = AlimKafka.class.getClassLoader();

    private final static Logger logger = Logger.getLogger(AlimKafka.class);

    public static void main(String[] args)  throws InterruptedException, IntrospectionException {

        StringBuilder stats = new StringBuilder();

        Properties props = new Properties();

        Producer<String, JsonNode> producer = null;

        String brokers = Config.getProperty("bootstrap.servers");
        String topic = Config.getProperty("topic");
        String numPartitions = Config.getProperty("numPartitions");
        String replication = Config.getProperty("replication");
        int total = Integer.valueOf(Config.getProperty("stats.total"));

        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);

        //Create topic if not exist
        AdminClient adminClient = AdminClient.create(props);

        NewTopic newTopic = new NewTopic(topic, Integer.valueOf(numPartitions), Integer.valueOf(replication).shortValue());
        CreateTopicsResult createTopicsResult = adminClient.createTopics(Collections.singleton(newTopic));

        try {
            createTopicsResult.all().get();
            // real failure cause is wrapped inside the raised ExecutionException
        } catch (ExecutionException e) {
            if (e.getCause() instanceof TopicExistsException) {
                logger.info("Topic already exists (delete manually to change settings) !!");
            } else if (e.getCause() instanceof TimeoutException) {
                logger.error("Timeout !!");
            } else {
                logger.error("Error !!" + e.getCause());
            }
        } finally {
            adminClient.close();
        }

        try {
            props.put(ProducerConfig.CLIENT_ID_CONFIG, "KafkaProducer");
            props.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG,
                    StringSerializer.class.getName());
            props.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG,
                    JsonSerializer.class.getName());
            props.put(ProducerConfig.MAX_IN_FLIGHT_REQUESTS_PER_CONNECTION, 4);
            ObjectMapper mapper = new ObjectMapper();

            producer = new KafkaProducer<String, JsonNode>(props);
            List<Person> persons = null;

            File file = new File(classLoader.getResource("listPersons.json").getFile());

            long i = 0;

            persons = mapper.readValue(file, new TypeReference<List<Person>>() {
            });

            while (i < total) {

                LocalDateTime start = LocalDateTime.now();

                Producer<String, JsonNode> finalProducer = producer;
                long size = persons.parallelStream().map(person -> {
                    try {
                        finalProducer.send(new ProducerRecord<String, JsonNode>(topic, mapper.valueToTree(person)));
                    } catch (Exception e) {
                        logger.error("Error IO" + e.getMessage());
                    }
                    return null;
                }).count();

                i = i + size;
                LocalDateTime end = LocalDateTime.now();

                long millis = ChronoUnit.MILLIS.between(start, end);

                stats.append(persons.size()).append(" messages inserted in ").append(millis).append(" ms.").append(System.lineSeparator());
            }
        } catch (IOException e) {
            logger.error("Error IO" + e.getMessage());
        } finally {
            Thread.sleep(5000);
            logger.info(stats);
            MetricsProducerReporter metrics = new MetricsProducerReporter(producer);
            metrics.displayMetrics();
            producer.close();
        }
    }

}
